module basic_functions
  implicit none
contains
  function break_next_fibre_els(N,fibres,thresh,broken) result(x)
    integer, intent(in) :: N
    integer, dimension(N) :: fibres
    real*8, dimension(N), intent(in) :: thresh
    integer :: broken
    real*8 :: x  !Extension at which the fibre breaks
    !Internal variables
    real*8 :: low
    logical :: found
    integer :: i, ipos
    
    !Find the next fibre to break
    found = .false.
    low = 0d0
    ipos = 0
    do i = 1,N
       if (.not. found) then
          if (fibres(i) == 1) then
             low = thresh(i)
             ipos = i
             found = .true.
          endif
       else
          if (fibres(i) == 1 .and. thresh(i) < low) then
             low = thresh(i)
             ipos = i
          endif
       endif
    enddo
    
    !Break the next fibre
    fibres(ipos) = 0

    !Return the threshold at which the fibre is broken
    x = thresh(ipos)/(1d0+DBLE(broken)/DBLE(N-broken))
    !Update number of broken fibres
    broken = broken + 1
  end function break_next_fibre_els

!Break the next fibre in local load sharing
  function break_next_fibre_lls(N,fibres,thresh,hoshk,holes) result(x)
    integer, intent(in) :: N
    integer, dimension(N) :: fibres
    real*8, dimension(N,2) :: thresh  !Thresholds for each fibre
    integer, dimension(N,3) :: hoshk
    integer, dimension(N,2) :: holes
    real*8 :: x
    !Internal variables
    real*8 :: thresh_factor, low
    integer :: i, k, ij, ipos
    logical :: found
    
    !Update effective thresholds for each fibre
    do i = 1,N
       thresh_factor = 1d0
       do k = 1,2
          if (holes(i,k) .ne. 0) then
             ij = holes(i,k)
             thresh_factor = thresh_factor + DBLE(hoshk(ij,2))/DBLE(hoshk(ij,3))
          endif
       enddo
       thresh(i,2) = thresh(i,1)/thresh_factor
    enddo
    
    !Find the next fibre to break
    found = .false.
    low = 0d0
    ipos = 0
    do i = 1,N
       if (.not. found) then
          if (fibres(i) == 1) then
             low = thresh(i,2)
             ipos = i
             found = .true.
          endif
       else
          if (fibres(i) == 1 .and. thresh(i,2) < low) then
             low = thresh(i,2)
             ipos = i
          endif
       endif
    enddo
    
    !Break the next fibre
    fibres(ipos) = 0

    !Return x-value at which the fibre breaks
    x = thresh(ipos,2)
  end function break_next_fibre_lls
  
!The Hoshen-Kopelman algorithm to find holes + the neighbours
  subroutine hoshen_kopelman(N,fibres,hoshk,holes)
    integer, intent(in) :: N  !Dimension of grid
    integer, dimension(N), intent(in) :: fibres
    integer, dimension(N,3), intent(out) :: hoshk !Equivalence relation array
    integer, dimension(N,2), intent(out) :: holes !Neighbouring holes
    !Internal variables
    integer :: i, k, ij  !Iteration variables (array positions)
    integer :: right1, left1  !Positions of neighbouring fibres
    integer :: inext, iprev   !Posotions of neighbouring fibres
    integer :: Nf, N1
    logical :: already_found
    integer :: counter
    
    hoshk = 0
    do ij = 1,N
       hoshk(ij,1) = ij
    enddo
    
    do i = 1,N
       if (fibres(i) == 0) then
          !Initialize positions of fibres to the "right"
          ij = i
          !Dimension 1
          if (i == N) then
             right1 = ij + 1 - N
             inext = 1
          else
             right1 = ij + 1
             inext = i + 1
          endif
          
          Nf = hoshk(ij,1)
          do while (Nf .ne. hoshk(Nf,1))
             Nf = hoshk(Nf,1)
          enddo
          
          !If the one to the right in D1 is broken as well
          if (fibres(inext) == 0) then
             N1 = hoshk(right1,1)
             do while (N1 .ne. hoshk(N1,1))
                N1 = hoshk(N1,1)
             enddo
             if (Nf .ne. N1) then
                hoshk(MAX(Nf,N1),1) = MIN(Nf,N1)
             endif
          endif
       endif
    enddo

    !Clean up the hoshk array to give all equivalence classes the same values
    do ij = 1,N
       Nf = hoshk(ij,1)
       do while (Nf .ne. hoshk(Nf,1))
          Nf = hoshk(Nf,1)
       enddo
       hoshk(ij,1) = Nf
    enddo
    
    !Find the number of broken fibres in each hole and its number of neighbours
    holes = 0
    do i = 1,N
       ij = i

       if (fibres(i) == 0) then
          !If this is broken, add one to the counter for fibres in this hole
          hoshk(hoshk(ij,1),2) = hoshk(hoshk(ij,1),2) + 1
       else
          counter = 0
          !Find positions of neighbouring fibres:

          !Dimension 1
          if (i == N) then
             right1 = ij + 1 - N
             left1 = ij - 1
             inext = 1
             iprev = N - 1
          else if (i == 1) then
             right1 = ij + 1
             left1 = ij + N - 1
             inext = 2
             iprev = N
          else
             right1 = ij + 1
             left1 = ij - 1
             inext = i + 1
             iprev = i - 1
          endif

          !Chech the neighbours for broken fibres:

          !Right in D1
          if (fibres(inext) == 0) then
             hoshk(hoshk(right1,1),3) = hoshk(hoshk(right1,1),3) + 1
             holes(i,1) = hoshk(right1,1)
             counter = counter + 1
          endif
          
          !Left in D1
          already_found = .false.
          if (fibres(iprev) == 0) then
             if (counter == 0) then
                hoshk(hoshk(left1,1),3) = hoshk(hoshk(left1,1),3) + 1
                holes(i,1) = hoshk(left1,1)
                counter = counter + 1
             else
                do k = 1,counter
                   if (hoshk(left1,1) == holes(i,k)) then
                      already_found = .true.
                   endif
                enddo
                if (.not. already_found) then
                   hoshk(hoshk(left1,1),3) = hoshk(hoshk(left1,1),3) + 1
                   holes(i,counter+1) = hoshk(left1,1)
                   counter = counter + 1
                endif
             endif
          endif   
       endif
    enddo
  end subroutine hoshen_kopelman
end module basic_functions
