module simulations
use basic_functions
use probability_distributions
implicit none

contains
!Simulations with uniform distribution
!============================================================================
  subroutine els_sim_uniform(N,samples,xlow,file_x,file_excl,file_incl,file_hole)
    integer, intent(in) :: N, samples
    real*8, intent(in) :: xlow
    character(len=*) :: file_x, file_excl, file_incl
    character(len=*), optional :: file_hole
    !Internal variables
    real*8, dimension(N,samples) :: x
    real*8, dimension(N) :: average_x
    integer, dimension(N,samples) :: excl, incl
    real*8, dimension(N) :: average_excl, average_incl
    integer, dimension(:,:,:), allocatable :: hole
    real*8, dimension(:,:), allocatable :: average_hole
    integer :: i, j
    !----------------------------------

    !Allocate arrays for hole size distribution if necessary
    if (present(file_hole)) then
       allocate(hole(N,N,samples))
       allocate(average_hole(N,N))
    endif

    !$OMP PARALLEL DO
    do i = 1,samples
       if (present(file_hole)) then
          call els_uniform(N,xlow,x(:,i),excl(:,i),incl(:,i),hole(:,:,i))
       else
          call els_uniform(N,xlow,x(:,i),excl(:,i),incl(:,i))
       endif
       write(*,*) 'Finished global load sharing number', i, ' of', samples
    enddo
    !$OMP END PARALLEL DO
    
    !Calculate averages over all samples
    do i = 1,N
       average_x(i) = SUM(x(i,:))/DBLE(samples)
       average_excl(i) = DBLE(SUM(excl(i,:)))/DBLE(samples)
       average_incl(i) = DBLE(SUM(incl(i,:)))/DBLE(samples)
    enddo
    if (present(file_hole)) then
       do j = 1,N
          do i = 1,N
             average_hole(i,j) = DBLE(SUM(hole(i,j,:)))/DBLE(samples)
             average_hole(i,j) = DBLE(i)/DBLE(j)*average_hole(i,j)
          enddo
       enddo
       deallocate(hole)
    endif

    !Write data to file
    open(1,file=file_x,status='unknown')
    do i = 1,N
       write(1,*) DBLE(i)/DBLE(N), average_x(i)
    enddo
    close(1)
    
    open(2,file=file_excl,status='unknown')
    do i = 1,N
       write(2,*) average_excl(i)
    enddo
    close(2)

    open(3,file=file_incl,status='unknown')
    do i = 1,N
       write(3,*) average_incl(i)
    enddo
    close(3)

    if (present(file_hole)) then
       open(4,file=file_hole,status='unknown')
       write(4,*) average_hole
       close(4)
       deallocate(average_hole)
    endif
  end subroutine els_sim_uniform

  subroutine lls_sim_uniform(N,samples,xlow,file_x,file_excl,file_incl,file_hole)
    integer, intent(in) :: N, samples
    real*8, intent(in) :: xlow
    character(len=*) :: file_x, file_excl, file_incl
    character(len=*), optional :: file_hole
    !Internal variables
    real*8, dimension(N,samples) :: x
    real*8, dimension(N) :: average_x
    integer, dimension(N,samples) :: excl, incl
    real*8, dimension(N) :: average_excl, average_incl
    integer, dimension(:,:,:), allocatable :: hole
    real*8, dimension(:,:), allocatable :: average_hole
    integer :: i, j
    !----------------------------------

    !Allocate arrays for hole size distribution if necessary
    if (present(file_hole)) then
       allocate(hole(N,N,samples))
       allocate(average_hole(N,N))
    endif

    !$OMP PARALLEL DO
    do i = 1,samples
       if (present(file_hole)) then
          call lls_uniform(N,xlow,x(:,i),excl(:,i),incl(:,i),hole(:,:,i))
       else
          call lls_uniform(N,xlow,x(:,i),excl(:,i),incl(:,i))
       endif
       write(*,*) 'Finished local load sharing number', i, ' of', samples
    enddo
    !$OMP END PARALLEL DO
    
    !Calculate averages over all samples
    do i = 1,N
       average_x(i) = SUM(x(i,:))/DBLE(samples)
       average_excl(i) = DBLE(SUM(excl(i,:)))/DBLE(samples)
       average_incl(i) = DBLE(SUM(incl(i,:)))/DBLE(samples)
    enddo
    if (present(file_hole)) then
       do j = 1,N
          do i = 1,N
             average_hole(i,j) = DBLE(SUM(hole(i,j,:)))/DBLE(samples)
             average_hole(i,j) = DBLE(i)/DBLE(j)*average_hole(i,j)
          enddo
       enddo
       deallocate(hole)
    endif

    !Write data to file
    open(1,file=file_x,status='unknown')
    do i = 1,N
       write(1,*) DBLE(i)/DBLE(N), average_x(i)
    enddo
    close(1)
    
    open(2,file=file_excl,status='unknown')
    do i = 1,N
       write(2,*) average_excl(i)
    enddo
    close(2)

    open(3,file=file_incl,status='unknown')
    do i = 1,N
       write(3,*) average_incl(i)
    enddo
    close(3)

    if (present(file_hole)) then
       open(4,file=file_hole,status='unknown')
       write(4,*) average_hole
       close(4)
       deallocate(average_hole)
    endif
  end subroutine lls_sim_uniform

  subroutine els_uniform(N,xlow,x,excl,incl,hole_dist)
    integer, intent(in) :: N
    real*8, intent(in) :: xlow
    real*8, dimension(N) :: x
    integer, dimension(N), intent(out) :: excl, incl
    integer, dimension(N,N), optional, intent(out) :: hole_dist
    !Internal variables
    real*8, dimension(N) :: thresh
    integer, dimension(N) :: fibres
    integer, dimension(:,:), allocatable :: hoshk
    integer, dimension(:,:), allocatable :: holes
    integer :: broken, size, next, z, i
    real*8 :: xmax
    !-----------------------------------

    fibres = 1
    broken = 0
    excl = 0
    incl = 0
    if (present(hole_dist)) then
       hole_dist = 0
       allocate(hoshk(N,3))
       allocate(holes(N,2))
    endif
    call random_number(thresh)
    call uniform(N,xlow,thresh)

    do z = 1,N
       !Find hole distribution
       if (present(hole_dist) .and. z > 1) then
          call hoshen_kopelman(N,fibres,hoshk,holes)
          do i = 1,N
             if (fibres(i) == 0 .and. hoshk(i,1) == i) then
                hole_dist(hoshk(i,2),z-1) = hole_dist(hoshk(i,2),z-1) + 1
             endif
          enddo
       endif
       !Break next fibre
       x(z) = break_next_fibre_els(N,fibres,thresh,broken)
    enddo
    if (present(hole_dist)) then
       hole_dist(N,N) = 1
       deallocate(hoshk,holes)
    endif

    !Locate bursts
    xmax = x(1)
    size = 0
    !Exclusive bursts
    do i = 1,N
       if (x(i) > xmax) then
          excl(size) = excl(size) + 1
          size = 0
          xmax = x(i)
       endif
       size = size + 1
    enddo
    excl(size) = excl(size) + 1

    !Inclusive bursts
    do i = 1,N-1
       size = 1
       next = i + 1
       do while ((x(next) .le. x(i)) .and. (next .ne. N))
          next = next + 1
          size = size + 1
       enddo
       if ((next == N) .and. (x(N) .le. x(i))) then
          size = size + 1
       endif
       incl(size) = incl(size) + 1
    enddo
    incl(1) = incl(1) + 1    !Add the last inclusive burst at i = N
  end subroutine els_uniform

  subroutine lls_uniform(N,xlow,x,excl,incl,hole_dist)
    integer, intent(in) :: N
    real*8, intent(in) :: xlow
    real*8, dimension(N) :: x
    integer, dimension(N), intent(out) :: excl, incl
    integer, dimension(N,N), optional, intent(out) :: hole_dist
    !Internal variables
    real*8, dimension(N,2) :: thresh
    integer, dimension(N) :: fibres
    integer, dimension(N,3) :: hoshk
    integer, dimension(N,2) :: holes
    integer :: size, next, z, i
    real*8 :: xmax
    !-----------------------------------

    fibres = 1
    excl = 0
    incl = 0
    if (present(hole_dist)) then
       hole_dist = 0
    endif
    call random_number(thresh(:,1))
    call uniform(N,xlow,thresh(:,1))

    do z = 1,N
       call hoshen_kopelman(N,fibres,hoshk,holes)
       !Find hole distribution
       if (present(hole_dist) .and. z > 1) then
          do i = 1,N
             if (fibres(i) == 0 .and. hoshk(i,1) == i) then
                hole_dist(hoshk(i,2),z-1) = hole_dist(hoshk(i,2),z-1) + 1
             endif
          enddo
       endif
       !Break next fibre
       x(z) = break_next_fibre_lls(N,fibres,thresh,hoshk,holes)
    enddo
    if (present(hole_dist)) then
       hole_dist(N,N) = 1
    endif

    !Locate bursts
    xmax = x(1)
    size = 0
    !Exclusive bursts
    do i = 1,N
       if (x(i) > xmax) then
          excl(size) = excl(size) + 1
          size = 0
          xmax = x(i)
       endif
       size = size + 1
    enddo
    excl(size) = excl(size) + 1

    !Inclusive bursts
    do i = 1,N-1
       size = 1
       next = i + 1
       do while ((x(next) .le. x(i)) .and. (next .ne. N))
          next = next + 1
          size = size + 1
       enddo
       if ((next == N) .and. (x(N) .le. x(i))) then
          size = size + 1
       endif
       incl(size) = incl(size) + 1
    enddo
    incl(1) = incl(1) + 1    !Add the last inclusive burst at i = N
  end subroutine lls_uniform


!Simulations with exponential distribution
!=============================================================================
  subroutine els_sim_exponential(N,samples,xlow,file_x,file_excl,file_incl,file_hole)
    integer, intent(in) :: N, samples
    real*8, intent(in) :: xlow
    character(len=*) :: file_x, file_excl, file_incl
    character(len=*), optional :: file_hole
    !Internal variables
    real*8, dimension(N,samples) :: x
    real*8, dimension(N) :: average_x
    integer, dimension(N,samples) :: excl, incl
    real*8, dimension(N) :: average_excl, average_incl
    integer, dimension(:,:,:), allocatable :: hole
    real*8, dimension(:,:), allocatable :: average_hole
    integer :: i, j
    !----------------------------------

    !Allocate arrays for hole size distribution if necessary
    if (present(file_hole)) then
       allocate(hole(N,N,samples))
       allocate(average_hole(N,N))
    endif

    !$OMP PARALLEL DO
    do i = 1,samples
       if (present(file_hole)) then
          call els_exponential(N,xlow,x(:,i),excl(:,i),incl(:,i),hole(:,:,i))
       else
          call els_exponential(N,xlow,x(:,i),excl(:,i),incl(:,i))
       endif
       write(*,*) 'Finished global load sharing number', i, ' of', samples
    enddo
    !$OMP END PARALLEL DO
    
    !Calculate averages over all samples
    do i = 1,N
       average_x(i) = SUM(x(i,:))/DBLE(samples)
       average_excl(i) = DBLE(SUM(excl(i,:)))/DBLE(samples)
       average_incl(i) = DBLE(SUM(incl(i,:)))/DBLE(samples)
    enddo
    if (present(file_hole)) then
       do j = 1,N
          do i = 1,N
             average_hole(i,j) = DBLE(SUM(hole(i,j,:)))/DBLE(samples)
             average_hole(i,j) = DBLE(i)/DBLE(j)*average_hole(i,j)
          enddo
       enddo
       deallocate(hole)
    endif

    !Write data to file
    open(1,file=file_x,status='unknown')
    do i = 1,N
       write(1,*) DBLE(i)/DBLE(N), average_x(i)
    enddo
    close(1)
    
    open(2,file=file_excl,status='unknown')
    do i = 1,N
       write(2,*) average_excl(i)
    enddo
    close(2)

    open(3,file=file_incl,status='unknown')
    do i = 1,N
       write(3,*) average_incl(i)
    enddo
    close(3)

    if (present(file_hole)) then
       open(4,file=file_hole,status='unknown')
       write(4,*) average_hole
       close(4)
       deallocate(average_hole)
    endif
  end subroutine els_sim_exponential

  subroutine lls_sim_exponential(N,samples,xlow,file_x,file_excl,file_incl,file_hole)
    integer, intent(in) :: N, samples
    real*8, intent(in) :: xlow
    character(len=*) :: file_x, file_excl, file_incl
    character(len=*), optional :: file_hole
    !Internal variables
    real*8, dimension(N,samples) :: x
    real*8, dimension(N) :: average_x
    integer, dimension(N,samples) :: excl, incl
    real*8, dimension(N) :: average_excl, average_incl
    integer, dimension(:,:,:), allocatable :: hole
    real*8, dimension(:,:), allocatable :: average_hole
    integer :: i, j
    !----------------------------------

    !Allocate arrays for hole size distribution if necessary
    if (present(file_hole)) then
       allocate(hole(N,N,samples))
       allocate(average_hole(N,N))
    endif

    !$OMP PARALLEL DO
    do i = 1,samples
       if (present(file_hole)) then
          call lls_exponential(N,xlow,x(:,i),excl(:,i),incl(:,i),hole(:,:,i))
       else
          call lls_exponential(N,xlow,x(:,i),excl(:,i),incl(:,i))
       endif
       write(*,*) 'Finished local load sharing number', i, ' of', samples
    enddo
    !$OMP END PARALLEL DO
    
    !Calculate averages over all samples
    do i = 1,N
       average_x(i) = SUM(x(i,:))/DBLE(samples)
       average_excl(i) = DBLE(SUM(excl(i,:)))/DBLE(samples)
       average_incl(i) = DBLE(SUM(incl(i,:)))/DBLE(samples)
    enddo
    if (present(file_hole)) then
       do j = 1,N
          do i = 1,N
             average_hole(i,j) = DBLE(SUM(hole(i,j,:)))/DBLE(samples)
             average_hole(i,j) = DBLE(i)/DBLE(j)*average_hole(i,j)
          enddo
       enddo
       deallocate(hole)
    endif

    !Write data to file
    open(1,file=file_x,status='unknown')
    do i = 1,N
       write(1,*) DBLE(i)/DBLE(N), average_x(i)
    enddo
    close(1)
    
    open(2,file=file_excl,status='unknown')
    do i = 1,N
       write(2,*) average_excl(i)
    enddo
    close(2)

    open(3,file=file_incl,status='unknown')
    do i = 1,N
       write(3,*) average_incl(i)
    enddo
    close(3)

    if (present(file_hole)) then
       open(4,file=file_hole,status='unknown')
       write(4,*) average_hole
       close(4)
       deallocate(average_hole)
    endif
  end subroutine lls_sim_exponential

  subroutine els_exponential(N,xlow,x,excl,incl,hole_dist)
    integer, intent(in) :: N
    real*8, intent(in) :: xlow
    real*8, dimension(N) :: x
    integer, dimension(N), intent(out) :: excl, incl
    integer, dimension(N,N), optional, intent(out) :: hole_dist
    !Internal variables
    real*8, dimension(N) :: thresh
    integer, dimension(N) :: fibres
    integer, dimension(:,:), allocatable :: hoshk
    integer, dimension(:,:), allocatable :: holes
    integer :: broken, size, next, z, i
    real*8 :: xmax
    !-----------------------------------

    fibres = 1
    broken = 0
    excl = 0
    incl = 0
    if (present(hole_dist)) then
       hole_dist = 0
       allocate(hoshk(N,3))
       allocate(holes(N,2))
    endif
    call random_number(thresh)
    call exponential(N,xlow,thresh)

    do z = 1,N
       !Find hole distribution
       if (present(hole_dist) .and. z > 1) then
          call hoshen_kopelman(N,fibres,hoshk,holes)
          do i = 1,N
             if (fibres(i) == 0 .and. hoshk(i,1) == i) then
                hole_dist(hoshk(i,2),z-1) = hole_dist(hoshk(i,2),z-1) + 1
             endif
          enddo
       endif
       !Break next fibre
       x(z) = break_next_fibre_els(N,fibres,thresh,broken)
    enddo
    if (present(hole_dist)) then
       hole_dist(N,N) = 1
       deallocate(hoshk,holes)
    endif

    !Locate bursts
    xmax = x(1)
    size = 0
    !Exclusive bursts
    do i = 1,N
       if (x(i) > xmax) then
          excl(size) = excl(size) + 1
          size = 0
          xmax = x(i)
       endif
       size = size + 1
    enddo
    excl(size) = excl(size) + 1

    !Inclusive bursts
    do i = 1,N-1
       size = 1
       next = i + 1
       do while ((x(next) .le. x(i)) .and. (next .ne. N))
          next = next + 1
          size = size + 1
       enddo
       if ((next == N) .and. (x(N) .le. x(i))) then
          size = size + 1
       endif
       incl(size) = incl(size) + 1
    enddo
    incl(1) = incl(1) + 1    !Add the last inclusive burst at i = N
  end subroutine els_exponential

  subroutine lls_exponential(N,xlow,x,excl,incl,hole_dist)
    integer, intent(in) :: N
    real*8, intent(in) :: xlow
    real*8, dimension(N) :: x
    integer, dimension(N), intent(out) :: excl, incl
    integer, dimension(N,N), optional, intent(out) :: hole_dist
    !Internal variables
    real*8, dimension(N,2) :: thresh
    integer, dimension(N) :: fibres
    integer, dimension(N,3) :: hoshk
    integer, dimension(N,2) :: holes
    integer :: size, next, z, i
    real*8 :: xmax
    !-----------------------------------

    fibres = 1
    excl = 0
    incl = 0
    if (present(hole_dist)) then
       hole_dist = 0
    endif
    call random_number(thresh(:,1))
    call exponential(N,xlow,thresh(:,1))

    do z = 1,N
       call hoshen_kopelman(N,fibres,hoshk,holes)
       !Find hole distribution
       if (present(hole_dist) .and. z > 1) then
          do i = 1,N
             if (fibres(i) == 0 .and. hoshk(i,1) == i) then
                hole_dist(hoshk(i,2),z-1) = hole_dist(hoshk(i,2),z-1) + 1
             endif
          enddo
       endif
       !Break next fibre
       x(z) = break_next_fibre_lls(N,fibres,thresh,hoshk,holes)
    enddo
    if (present(hole_dist)) then
       hole_dist(N,N) = 1
    endif

    !Locate bursts
    xmax = x(1)
    size = 0
    !Exclusive bursts
    do i = 1,N
       if (x(i) > xmax) then
          excl(size) = excl(size) + 1
          size = 0
          xmax = x(i)
       endif
       size = size + 1
    enddo
    excl(size) = excl(size) + 1

    !Inclusive bursts
    do i = 1,N-1
       size = 1
       next = i + 1
       do while ((x(next) .le. x(i)) .and. (next .ne. N))
          next = next + 1
          size = size + 1
       enddo
       if ((next == N) .and. (x(N) .le. x(i))) then
          size = size + 1
       endif
       incl(size) = incl(size) + 1
    enddo
    incl(1) = incl(1) + 1    !Add the last inclusive burst at i = N
  end subroutine lls_exponential


!==========================================================================
! Initialize random_number
  subroutine init_seed(N,a)
    integer, intent(in) :: N
    real*8, intent(in) :: a
    !Internal variables
    integer, dimension(N) :: seed
    real*8, dimension(N) :: rand
    integer :: i
    
    seed = 9
    call random_number(rand)
    do i = 1,N
       seed(i) = int(a*rand(i)) + 1
    enddo
    call random_seed(put=seed)
  end subroutine init_seed
end module simulations
