module basic_functions
  implicit none
contains
  function break_next_fibre_els(N,fibres,thresh,broken) result(x)
    integer, intent(in) :: N
    integer, dimension(N,N) :: fibres
    real*8, dimension(N,N), intent(in) :: thresh
    integer :: broken
    real*8 :: x  !Extension at which the fibre breaks
    !Internal variables
    real*8 :: low
    logical :: found
    integer :: i, j, ipos, jpos
    
    !Find the next fibre to break
    found = .false.
    low = 0d0
    ipos = 0
    jpos = 0
    do j = 1,N
       do i = 1,N
          if (.not. found) then
             if (fibres(i,j) == 1) then
                low = thresh(i,j)
                ipos = i
                jpos = j
                found = .true.
             endif
          else
             if (fibres(i,j) == 1 .and. thresh(i,j) < low) then
                low = thresh(i,j)
                ipos = i
                jpos = j
             endif
          endif
       enddo
    enddo
    
    !Break the next fibre
    fibres(ipos,jpos) = 0

    !Return the threshold at which the fibre is broken
    x = thresh(ipos,jpos)/(1d0+DBLE(broken)/DBLE(N**2-broken))
    !Update number of broken fibres
    broken = broken + 1
  end function break_next_fibre_els

!Break the next fibre in local load sharing
  function break_next_fibre_lls(N,fibres,thresh,hoshk,holes) result(x)
    integer, intent(in) :: N
    integer, dimension(N,N) :: fibres
    real*8, dimension(N,N,2) :: thresh  !Thresholds for each fibre
    integer, dimension(N**2,3) :: hoshk
    integer, dimension(N,N,4) :: holes
    real*8 :: x
    !Internal variables
    real*8 :: thresh_factor, low
    integer :: i, j, k, ij, ipos, jpos
    logical :: found
    
    !Update effective thresholds for each fibre
    do j = 1,N
       do i = 1,N
          thresh_factor = 1d0
          do k = 1,4
             if (holes(i,j,k) .ne. 0) then
                ij = holes(i,j,k)
                thresh_factor = thresh_factor + DBLE(hoshk(ij,2))/DBLE(hoshk(ij,3))
             endif
          enddo
          thresh(i,j,2) = thresh(i,j,1)/thresh_factor
       enddo
    enddo
    
    !Find the next fibre to break
    found = .false.
    low = 0d0
    ipos = 0
    jpos = 0
    do j = 1,N
       do i = 1,N
          if (.not. found) then
             if (fibres(i,j) == 1) then
                low = thresh(i,j,2)
                ipos = i
                jpos = j
                found = .true.
             endif
          else
             if (fibres(i,j) == 1 .and. thresh(i,j,2) < low) then
                low = thresh(i,j,2)
                ipos = i
                jpos = j
             endif
          endif
       enddo
    enddo
    
    !Break the next fibre
    fibres(ipos,jpos) = 0

    !Return x-value at which the fibre breaks
    x = thresh(ipos,jpos,2)
  end function break_next_fibre_lls
  
!The Hoshen-Kopelman algorithm to find holes + the neighbours
  subroutine hoshen_kopelman(N,fibres,hoshk,holes)
    integer, intent(in) :: N  !Dimension of grid
    integer, dimension(N,N), intent(in) :: fibres
    integer, dimension(N**2,3), intent(out) :: hoshk !Equivalence relation array
    integer, dimension(N,N,4), intent(out) :: holes  !Neighbouring holes
    !Internal variables
    integer :: i, j, k, ij  !Iteration variables (array positions)
    integer :: right1, left1, right2, left2  !Positions of neighbouring fibres
    integer :: inext, iprev, jnext, jprev !Positions of neighbouring fibres
    integer :: Nf, N1, N2
    logical :: already_found
    integer :: counter
    
    hoshk = 0
    do ij = 1,N**2
       hoshk(ij,1) = ij
    enddo
    
    do j = 1,N
       do i = 1,N
          if (fibres(i,j) == 0) then
             !Initialize positions of fibres to the "right"
             ij = i + (j-1)*N
             !Dimension 1
             if (i == N) then
                right1 = ij + 1 - N
                inext = 1
             else
                right1 = ij + 1
                inext = i + 1
             endif
             !Dimension 2
             if (j == N) then
                right2 = i
                jnext = 1
             else
                right2 = ij + N
                jnext = j + 1
             endif
             
             Nf = hoshk(ij,1)
             do while (Nf .ne. hoshk(Nf,1))
                Nf = hoshk(Nf,1)
             enddo
             
             !If the one to the right in D1 is broken as well
             if (fibres(inext,j) == 0) then
                N1 = hoshk(right1,1)
                do while (N1 .ne. hoshk(N1,1))
                   N1 = hoshk(N1,1)
                enddo
                if (Nf .ne. N1) then
                   hoshk(MAX(Nf,N1),1) = MIN(Nf,N1)
                endif
             endif

             !If the on to the right in D2 is broken as well
             if (fibres(i,jnext) == 0) then
                N2 = hoshk(right2,1)
                do while (N2 .ne. hoshk(N2,1))
                   N2 = hoshk(N2,1)
                enddo
                if (Nf .ne. N2) then
                   hoshk(MAX(Nf,N2),1) = MIN(Nf,N2)
                endif
             endif
          endif
       enddo
    enddo

    !Clean up the hoshk array to give all equivalence classes the same values
    do ij = 1,N**2
       Nf = hoshk(ij,1)
       do while (Nf .ne. hoshk(Nf,1))
          Nf = hoshk(Nf,1)
       enddo
       hoshk(ij,1) = Nf
    enddo
    
    !Find the number of broken fibres in each hole and its number of neighbours
    holes = 0
    do j = 1,N
       do i = 1,N
          ij = i + (j-1)*N

          if (fibres(i,j) == 0) then
             !If this is broken, add one to the counter for fibres in this hole
             hoshk(hoshk(ij,1),2) = hoshk(hoshk(ij,1),2) + 1
          else
             counter = 0
             !Find positions of neighbouring fibres:

             !Dimension 1
             if (i == N) then
                right1 = ij + 1 - N
                left1 = ij - 1
                inext = 1
                iprev = N - 1
             else if (i == 1) then
                right1 = ij + 1
                left1 = ij + N - 1
                inext = 2
                iprev = N
             else
                right1 = ij + 1
                left1 = ij - 1
                inext = i + 1
                iprev = i - 1
             endif

             !Dimension 2
             if (j == N) then
                right2 = i
                left2 = ij - N
                jnext = 1
                jprev = N - 1
             else if (j == 1) then
                right2 = ij + N
                left2 = i + (N - 1)*N
                jnext = 2
                jprev = N
             else
                right2 = ij + N
                left2 = ij - N
                jnext = j + 1
                jprev = j - 1
             endif

             !Chech the neighbours for broken fibres:

             !Right in D1
             if (fibres(inext,j) == 0) then
                hoshk(hoshk(right1,1),3) = hoshk(hoshk(right1,1),3) + 1
                holes(i,j,1) = hoshk(right1,1)
                counter = counter + 1
             endif
             
             !Left in D1
             already_found = .false.
             if (fibres(iprev,j) == 0) then
                if (counter == 0) then
                   hoshk(hoshk(left1,1),3) = hoshk(hoshk(left1,1),3) + 1
                   holes(i,j,1) = hoshk(left1,1)
                   counter = counter + 1
                else
                   do k = 1,counter
                      if (hoshk(left1,1) == holes(i,j,k)) then
                         already_found = .true.
                      endif
                   enddo
                   if (.not. already_found) then
                      hoshk(hoshk(left1,1),3) = hoshk(hoshk(left1,1),3) + 1
                      holes(i,j,counter+1) = hoshk(left1,1)
                      counter = counter + 1
                   endif
                endif
             endif
             
             !Right in D2
             already_found = .false.
             if (fibres(i,jnext) == 0) then
                if (counter == 0) then
                   hoshk(hoshk(right2,1),3) = hoshk(hoshk(right2,1),3) + 1
                   holes(i,j,1) = hoshk(right2,1)
                   counter = counter + 1
                else
                   do k = 1,counter
                      if (hoshk(right2,1) == holes(i,j,k)) then
                         already_found = .true.
                      endif
                   enddo
                   if (.not. already_found) then
                      hoshk(hoshk(right2,1),3) = hoshk(hoshk(right2,1),3) + 1
                      holes(i,j,counter+1) = hoshk(right2,1)
                      counter = counter + 1
                   endif
                endif
             endif
             
             !Left in D2
             already_found = .false.
             if (fibres(i,jprev) == 0) then
                if (counter == 0) then
                   hoshk(hoshk(left2,1),3) = hoshk(hoshk(left2,1),3) + 1
                   holes(i,j,1) = hoshk(left2,1)
                   counter = counter + 1
                else
                   do k = 1,counter
                      if (hoshk(left2,1) == holes(i,j,k)) then
                         already_found = .true.
                      endif
                   enddo
                   if (.not. already_found) then
                      hoshk(hoshk(left2,1),3) = hoshk(hoshk(left2,1),3) + 1
                      holes(i,j,counter+1) = hoshk(left2,1)
                      counter = counter + 1
                   endif
                endif
             endif

          endif
       enddo
    enddo
  end subroutine hoshen_kopelman
end module basic_functions
